#!/usr/bin/perl
# 
# sb::Plugin::Podcaster - Module for sb
# == written by T.Otani <ootani@segausers.gr.jp> ===
# == Copyright (C) 2004 SimpleBoxes/SerendipityNZ ==

# 0.04 [2006/11/06] changed _entry/_generate_link/_get_sound to handle pdf
# 0.03 [2006/10/27] changed how to check the version
# 0.02 [2006/08/03] changed detail to point new site address
# 0.01 [2006/02/21] renamed. Thanks Gin-Lime!
# 0.00 [2006/02/20] created

if ($sb::VERSION =~ /^2.(\d\d)[RD]$/i)
{
	die("This plugin works under Serene Bach 2.04R or later.\n") if ($1 < 4 || $1 > 49);
}
else
{
	die("This plugin works for Serene Bach.\n");
}

# ==================================================
package sb::Plugin::Podcaster;
# --------------------------------------------------
# // initialization for plugin
# --------------------------------------------------
use sb::Plugin ();
sb::Plugin->register_plugin(
	'lang' => {
		'ja' => 'euc',
		'en' => 'ascii',
	},
	'text' => {
		'type'    => 'admin',
		'name'    => 'Cast Master', # named by Gin-Lime
		'text'    => 'supports podcasting',
		'author'  => 'takkyun',
		'detail'  => 'http://serenebach.net/',
		'version' => '0.04',
	},
	'file' => 'podcaster.txt',
	'data' => 1,
);
sb::Plugin->register_admin_module(
	'mode'   => 'podcast',
	'level'  => 2,
	'module' => 'sb::Admin::Podcaster',
);
sb::Plugin->register_content_module(
	'type'     => 'main',
	'callback' => \&sb::Content::Podcaster::_common,
	'field'    => 'additional_feed',
);
# ==================================================
# --------------------------------------------------
# // override default subroutine
# --------------------------------------------------
use sb::App::Feed ();
use sb::Data::Image ();
my $original_feed_run    = \&sb::App::Feed::run;
my $original_extension   = \&sb::Data::Image::_file_extension;
my $original_contenttype = \&sb::Data::Image::_file_contenttype;
&_initialize_plugin(); # initialize to override default function
sub _initialize_plugin
{
	local $^W = 0; # ignore warning
	no strict 'refs';
	my $weblog = sb::Data->load('Weblog','id'=>0);
	my $flag = undef;
	if ($weblog)
	{
		my $plugin = $weblog->plugin;
		my $name = __FILE__;
		$name =~ s/.*[:\/\\](.*)/$1/;
		$flag = ( index($plugin,'[' . $name . ']') > -1 );
	}
	if ($flag)
	{
		*sb::App::Feed::run = \&_feed_run;
		*sb::Data::Image::_file_extension = \&_file_extension;
		*sb::Data::Image::_file_contenttype = \&_file_contenttype;
	}
}
# ==================================================
# --------------------------------------------------
# // declaration for constant value
# --------------------------------------------------
sub TEMPLATE_RSS2  (){ 'template_rss2.xml' };
sub CONTENT_TYPE   (){ 'text/xml' };
sub OUTPUT_CHARSET (){ 'utf-8' };
sub OUTPUT_CODE    (){ 'utf8' };
sub FEED_NUM       (){ 10 };
sub TIME_FORMAT    (){ '%Week%, %Day% %MonShort% %Year% %Hour%:%Min%:%Sec% %Zone%' };
sub RSS2_ENCLOSURE (){ '<enclosure url="%s" length="%d" type="%s" />' };
sub RSS2_DURATION  (){ '<itunes:duration>%s</itunes:duration>' };
# --------------------------------------------------
# // subroutines
# --------------------------------------------------
sub _feed_run
{
	my $class = shift;
	my $cgi  = sb::Interface->get;
	my $type = $cgi->value('feed');
	return &{$original_feed_run}($class) if ($type ne 'rss2' and $type ne 'podcast'); # call original one
	my $self = sb::App::new($class,@_);
	my $conf = sb::Config->get;
	my $dir  = sb::Plugin->get_resource_dir;
	my $base = sb::TemplateManager->new($self->load_template('dir'=>$dir,'file'=>TEMPLATE_RSS2));
	my ($mode,$id,$cond) = &sb::App::Feed::_check_mode($cgi,$conf);
	if ($type eq 'podcast')
	{ # extract podcasting entry
		$cond = '<!--[podcast]';
		$mode = 'srch';
	}
	if ($base ne '')
	{
		$base = sb::Content->output($base,
			'mode' => $mode,
			'id'   => $id,
			'page' => int( $cgi->value('page') ),
			'cond' => $cond,
			'time' => $self->{'time'},
			'num'  => FEED_NUM,
			'extend' => {
				'main' => {
					'_main'      => \&_common_parts,
					'feed_entry' => \&_entry,
				},
			},
		);
		my $lang = sb::Language->get;
		if ($lang->charcode ne OUTPUT_CODE)
		{ # convert character code
			$lang->checkcode('',$lang->charcode);
			$base = $lang->convert($base,OUTPUT_CODE);
		}
	}
	print $cgi->head('type'=>CONTENT_TYPE,'charset'=>OUTPUT_CHARSET) . $base;
}
sub _common_parts
{
	my $cms = shift;
	my %var = @_;
	sb::Content::_common_parts($cms,%var);
	$cms->num(0);
	$cms->tag('feed_site_encoding'=>OUTPUT_CHARSET);
	$cms->tag('feed_date'=>sb::Time->format('time'=>$var{'time'},'form'=>TIME_FORMAT));
	my $data = sb::Plugin->get_data;
	$cms->tag('feed_owner'=>$data->data);
	if ($data->setting ne '')
	{
		$cms->tag('feed_site_image'=>$data->setting);
		$cms->block('rss2_image'=>1);
	}
}
sub _entry
{
	my $cms = shift;
	my %var = @_;
	my $num = 0;
	foreach my $entry ( @{$var{'entry'}} )
	{
		my $date = sb::Time->format('time'=>$entry->date,'form'=>TIME_FORMAT,'zone'=>$entry->tz);
		my $body = $entry->body;
		$cms->num($num);
		$cms->tag('feed_entry_url'=>=>$entry->permalink('cat'=>$var{'cat'}));
		$cms->tag('feed_entry_title'=>$entry->subj);
		$cms->tag('feed_entry_summary'=>sb::Text->entitize($entry->sum));
		$cms->tag('feed_entry_date'=>$date);
		$cms->tag('feed_entry_author'=>$entry->authname($var{'user'}));
		$cms->tag('feed_entry_keyword'=>$entry->key);
		if ($body =~ /<\!--\[podcast\](\d+?)\[podcast\]-->/)
		{
			my $obj = sb::Data->load('Image','id'=>$1);
			if ($obj and ($obj->file =~ /\.mp[34]$/ or $obj->file =~ /\.m4a$/ or $obj->file =~ /\.pdf$/))
			{
				my $file = $obj->get_path;
				my $info = Sound::Info->new($file);
				my $podcast = sprintf(RSS2_ENCLOSURE,$obj->get_url,(-s $file),$obj->get_content_type);
				$cms->tag('feed_entry_enclosure'=>$podcast);
				$cms->tag('feed_entry_duration'=>sprintf(RSS2_DURATION,$info->duration));
			}
		}
		$num++;
	}
	return($num);
}
sub _file_extension
{
	my $content_type = shift;
	return('.mp3') if ($content_type eq 'audio/mpeg');
	return('.m4a') if ($content_type eq 'audio/x-m4a');
	return('.mp4') if ($content_type eq 'video/mp4');
	return &{$original_extension}($content_type);
}
sub _file_contenttype
{
	my $name = shift;
	return('audio/mpeg')  if ($name =~ /\.mp3$/i);
	return('audio/x-m4a') if ($name =~ /\.m4a$/i);
	return('video/mp4')   if ($name =~ /\.mp4$/i);
	return &{$original_contenttype}($name);
}

# ==================================================
package sb::Content::Podcaster;
sub _common
{
	my $cms = shift;
	my %var = @_;
	my $path = $var{'conf'}->value('conf_srv_cgi') . $var{'conf'}->value('basic_feed');
	$cms->num(0);
	$cms->tag('site_rss2'=>$path . 'rss2');
	$cms->tag('site_podcast'=>$path . 'podcast');
	return 1;
}

# ==================================================
package sb::Admin::Podcaster;
# --------------------------------------------------
# // declaration for global variables
# --------------------------------------------------
use strict;
use vars qw( @ISA );
# --------------------------------------------------
# // configuration for inheritance / dependancy
# --------------------------------------------------
use sb::Config ();
use sb::Language ();
use sb::Interface ();
use sb::Admin::Upload ();
@ISA = qw( sb::Admin::Upload );
# --------------------------------------------------
# // declaration for constant value
# --------------------------------------------------
sub TEMPLATE     (){ 'podcaster.html' };
sub DEFAULT_TYPE (){ 'main' };
sub THUMB_SIZE   (){ 80 };
sub TABLE_WIDTH  (){ 495 };
sub LIST_COLUMN  (){ 7 };
sub THUMB_PREFIX (){ 'art' };
sub DEWPLAYER    (){ 'dewplayer.swf' };
sub DEFAULT_TEMP (){ "<div class=\"podcast\">%Image%<br />%Player% <a href=\"%FilePath%\">[Download]</a></div>" };
# --------------------------------------------------
# // public functions - callback
# --------------------------------------------------
sub callback
{ # callbacks
	my $self = shift;
	$self->{'data'} = sb::Plugin->get_data;
	$self->_initialize;
	return ( $self->{'regi'} ) 
		? $self->_update_podcast(@_)
		: $self->_open_podcast(@_);
}
# --------------------------------------------------
# // private functions - main routine
# --------------------------------------------------
sub _update_podcast
{
	my $self = shift;
	my %param = ('message'=>'',@_);
	my $lang = sb::Language->get;
	my $cgi = sb::Interface->get;
	my $msg = '';
	SWITCH_REGI: {
		$_ = $cgi->value('__regi');
		/^upload$/ && do
		{ # upload file
			my $num = $self->upload_image('over'=>($cgi->value('upload_overwrite') eq 'on'));
			if ($num > 0)
			{
				$self->_add_thumb(sb::Data->load('Image','sort'=>'date','order'=>1,'num'=>1));
				$msg = sprintf($lang->string('parts_add_comp'),$num);
			}
			else
			{
				$msg = $lang->string('error_failtoadd');
			}
			last SWITCH_REGI;
		}; # end of upload
		/^casting$/ && do
		{ # write an entry for podcast
			last SWITCH_REGI if ($cgi->value('id') eq '');
			my $obj = sb::Data->load('Image','id'=>$cgi->value('id'));
			last SWITCH_REGI if (!$obj);
			$self->{'mode'} = 'new';
			return $self->_open_entry('newtext'=>$self->_generate_link($obj));
		}; # end of casting
		/^update$/ && do
		{ # rename or do an action
			my @objs = ();
			if ($cgi->value('namechange') ne '')
			{ # rename
				my @ids = split("\0",$cgi->value('iid'));
				@objs = sb::Data->load('Image','cond'=>{'id'=>\@ids});
				foreach my $obj (@objs)
				{
					my $name = $cgi->value('img_name' . $obj->id);
					$obj->name(sb::Text->entitize($name));
				}
				$msg = $lang->string('parts_editcomp');
			}
			else
			{ # do an action
				my @sels = split("\0",$cgi->value('sel'));
				@objs = sb::Data->load('Image','cond'=>{'id'=>\@sels});
				if ($cgi->value('regi_action') eq 'del')
				{ # delete
					foreach my $obj (@objs)
					{
						$obj->erase;
					}
					$msg = ($#objs + 1) . $lang->string('parts_deleted');
				}
				elsif ($cgi->value('regi_action') =~ /^\[(.+)\]$/)
				{ # replace
					my $new = $1;
					foreach my $obj (@objs)
					{
						$obj->rename_file('dir'=>$new);
					}
					$msg = $lang->string('parts_editcomp');
				}
			} # end of updating process
			sb::Data->update(@objs) if (@objs);
			last SWITCH_REGI;
		}; # end of update
		/^info$/ && do { # update podcast information
			$self->{'data'}->data(sb::Text->entitize($cgi->value('podcast_author')));
			$self->{'data'}->setting($cgi->value('podcast_image'));
			$self->{'data'}->text($cgi->value('podcast_template'));
			sb::Plugin->set_data('data'=>$self->{'data'});
			$msg = $lang->string('parts_editcomp');
			last SWITCH_REGI;
		}; # end of info
	};
	$msg = $lang->string('error_unknown') if ($msg eq '');
	return $self->_open_podcast('message'=>$msg);
}
sub _open_podcast
{
	my $self = shift;
	my %param = ('message'=>'',@_);
	my $cgi = sb::Interface->get;
	my $cms = sb::TemplateManager->new(sb::Plugin->load_template('file'=>TEMPLATE));
	my $type = $cgi->value('__type') || DEFAULT_TYPE;
	SWITCH_TYPE: {
		$_ = $type;
		/^copy$/ && do
		{
			# nothing to do
			last SWITCH_TYPE;
		}; # end of copy
		/^info$/ && do
		{
			my $icon = sb::Config->get->value('srv_temp') . &sb::Data::Image::PARTS_DIR();
			$icon .= &sb::Data::Image::ICON_UNDEFINED();
			$cms->num(0);
			$cms->tag('sb_podcast_author'=>$self->{'data'}->data);
			if ($self->{'data'}->setting)
			{
				$cms->tag('sb_podcast_preview'=>$self->{'data'}->setting);
			}
			else
			{
				$cms->tag('sb_podcast_preview'=>$icon);
			}
			$cms->tag('sb_podcast_default_icon'=>$icon);
			my $temp = $self->{'data'}->text;
			$temp = DEFAULT_TEMP if ($temp eq '');
			$cms->tag('sb_podcast_template'=>sb::Text->entitize($temp));
			my $selector = '';
			my @imgs = sb::Data->load('Image','sort'=>'date','order'=>1);
			for (my $i=0;$i<@imgs;$i++)
			{
				my $img = $imgs[$i];
				next if (!$img->is_image);
				$selector .= '<option value="' . $img->get_url . '"';
				$selector .= ' selected="selected"' if ($img->get_url eq $self->{'data'}->setting);
				$selector .= '>' . $img->name . '</option>';
			}
			$cms->tag('sb_podcast_image'=>$selector);
			last SWITCH_TYPE;
		}; # end of info
		/^main$/ && do
		{
			my $num  = int($self->setup_list->{'dispnum'} / 2);
			my $page = int($cgi->value('page'));
			$page = int($cgi->value('now_page')) + 1 if ( $cgi->value('next') ne '' );
			$page = int($cgi->value('now_page')) - 1 if ( $cgi->value('prev') ne '' );
			my @files = $self->_get_sound($page * $num,($page + 1) * $num);
			$cms->num(0);
			$cms->tag('sb_list_page'=>$self->display_pagelink( # page link
					'mode'    => 'podcast',
					'column'  => LIST_COLUMN,
					'all'     => $self->{'matched_sound'},
					'printed' => $#files + 1,
					'num'     => $num,
					'params'  => [],
				)
			);
			$self->imagedir_selector( # image selector for selection
				'cms'    => $cms,
				'tag'    => 'sb_img_dirsel',
				'select' => '',
				'format' => '[%s]',
			);
			$self->listmain(
				'template' => $cms,
				'block'    => 'sb_podcast_list',
				'objects'  => \@files,
				'tags'     => {
					'sb_img_id'        => 'id',
					'sb_img_image'     => \&_display_image,
					'sb_img_nametext'  => 'name',
					'sb_img_disable'   => \&sb::Admin::Upload::_is_editable,
					'sb_img_dirdisp'   => 'dir',
					'sb_img_imagsize'  => \&_display_size,
					'sb_img_statclass' => \&sb::Admin::Upload::_display_status,
					'sb_img_sel'       => \&sb::Admin::Upload::_display_checkbox,
					'sb_img_player'    => \&_display_player,
					'sb_img_height'    => sub { THUMB_SIZE },
					'sb_site_cgi'      => sub { return $self->get_script_path; },
				},
			);
			$self->imagedir_selector( # image selector for uploda form
				'cms'   => $cms,
				'tag'   => 'sb_img_dir',
				'over'  => 'check',
			);
			last SWITCH_TYPE;
		}; # end of main
	}; # end of SWITCH_TYPE
	$cms->num(0);
	$cms->tag('sb_podcast_menu_' . $type => 'class="current"');
	$cms->block('sb_podcast_' . $type => 1);
	if ($param{'message'} ne '')
	{
		$cms->num(0);
		$cms->tag('sb_process_message'=>$param{'message'});
		$cms->block('sb_process_message'=>1);
	}
	$self->common_template_parts($cms);
	return sb::Interface->get->head('type'=>'text/html') . $self->set_main($cms->output);
}
# --------------------------------------------------
# // private functions - utilities
# --------------------------------------------------
sub _initialize
{
	my $self = shift;
	my $check = $self->{'data'}->extra;
	my $buf;
	return if ($check);
	my $conf = sb::Config->get;
	my $orig = sb::Plugin->get_resource_dir . DEWPLAYER;
	my $dest = $conf->value('conf_dir_base') . $conf->value('dir_style') . DEWPLAYER;
	open(ORIG,"<$orig");
	open(DEST,">$dest");
	binmode(ORIG);
	binmode(DEST);
	while ( read(ORIG,$buf,1024) )
	{
		print DEST $buf;
	}
	close(ORIG);
	close(DEST);
	chmod($conf->value('basic_file_attr'),$dest);
	$self->{'data'}->extra('done');
	sb::Plugin->set_data('data'=>$self->{'data'});
}
sub _get_player
{
	my $self = shift;
	my $target = shift;
	return('') if (!$self->{'data'}->extra);
	my $conf = sb::Config->get;
	my $path = $conf->value('conf_dir_base') . $conf->value('dir_style') . DEWPLAYER;
	return('') if (!-e $path);
	return('') if ($target !~ /\.mp3$/);
	my $url = $conf->value('conf_srv_base') . $conf->value('dir_style') . DEWPLAYER;
	$url .= '?mp3=' . $target;
	return <<"__DEWPLAYER__";
<object type=\"application/x-shockwave-flash\" data=\"$url\" width=\"200\" height=\"20\">
<param name=\"movie\" value=\"$url\" />
</object>
__DEWPLAYER__
}
sub _get_sound
{
	my $self = shift;
	my ($bgn,$end) = @_;
	my $cnt = 0;
	my @all = sb::Data->load('Image','sort'=>'date','order'=>1);
	my @out = ();
	for (my $i=0;$i<@all;$i++)
	{
		my $obj = $all[$i];
		if ($obj->file =~ /\.mp[34]$/ or $obj->file =~ /\.m4a$/ or $obj->file =~ /\.pdf$/)
		{
			push (@out,$obj) if ($bgn <= $cnt and $cnt < $end);
			$cnt++;
		}
	}
	$self->{'matched_sound'} = $cnt;
	return @out;
}
sub _display_image
{
	my $self = shift;
	my $obj  = shift;
	return &sb::Admin::Upload::_display_image($self,$obj) if ($obj->thumb eq '');
	my ($w,$h) = $self->_get_size($obj,'max_w'=>THUMB_SIZE,'max_h'=>THUMB_SIZE);
	my $image = '<img src="' . $obj->get_url('type'=>'thumb') . '"';
	$image .= 'alt="' . $obj->name . '" title="' . $obj->name . '" ';
	$image .= 'width="' . $w . '" height="' . $h . '" />';
	return $self->clip_text(
		'text' => $image,
		'base' => '?__mode=upload&amp;image_id=' . $obj->id,
		'user' => $obj->auth,
		'length' => length($image),
	);
}
sub _display_size
{
	my $self = shift;
	my $obj  = shift;
	my $info = Sound::Info->new($obj->get_path);
	my $size = sprintf("%.2f",$obj->get_filesize / 1024);
	return( $info->duration . '<br />' . $size . 'MB' );
}
sub _display_player
{
	my $self = shift;
	my $obj  = shift;
	return('') if ($obj->file !~ /\.mp3$/);
	return $self->_get_player($obj->get_url);
}
# --------------------------------------------------
# // private functions - sound object
# --------------------------------------------------
sub _add_thumb
{
	my $self = shift;
	my $obj = shift;
	return if (!$obj);
	return if ($obj->file !~ /\.mp[34]$/ and $obj->file !~ /\.m4a$/);
	my $info = Sound::Info->new($obj->get_path);
	if ($info->artwork)
	{
		my $path = sb::Config->get->value('conf_dir_base') . $obj->dir;
		my $name = $obj->file;
		$name =~ s/(.*)\.m(4a|p3)$/$1/;
		$name = THUMB_PREFIX . $obj->id . '_' . $name . '.' . $info->arttype;
		$path .= $name;
		eval
		{
			die('file exit') if (-e $path);
			open(BINOUT,">$path") or die('failed file open');
			binmode(BINOUT);
			print BINOUT $info->artwork;
			close(BINOUT);
			chmod(sb::Config->get->value('basic_file_attr'),$path);
		};
		return if ($@);
		$obj->thumb($name);
		sb::Data->update($obj);
	}
}
sub _get_size
{
	my $self = shift;
	my $obj  = shift;
	return(0,0) if (!$obj);
	return(0,0) if ($obj->thumb eq '');
	my %param = ('max_w'=>0,'max_h'=>0,@_);
	my ($w,$h) = (0,0);
	my $path = $obj->get_path('type'=>'thumb');
	# artwork which is included audio/video file supports png or jpg only.
	if ($path =~ /\.png$/)
	{
		($w,$h) = &sb::Data::Image::_get_size_png($path);
	}
	elsif ($path =~ /\.jpg$/)
	{
		($w,$h) = &sb::Data::Image::_get_size_jpg($path);
	}
	if ($param{'max_w'} and $param{'max_h'})
	{
		($w,$h) = &sb::Data::Image::_resize($w,$h,$param{'max_w'},$param{'max_h'});
	}
	return($w,$h);
}
sub _generate_link
{
	my $self = shift;
	my $obj  = shift;
	return('') if (!$obj);
	my $image = '';
	my $option = '';
	my $temp = $self->{'data'}->text;
	$temp = DEFAULT_TEMP if ($temp eq '');
	if ($obj->thumb eq '')
	{
		$image = $obj->get_as_tag('type'=>'file');
	}
	else
	{
		my ($w,$h) = $self->_get_size($obj);
		$image .= '<img src="' . $obj->get_url('type'=>'thumb') . '"';
		$image .= 'alt="' . $obj->name . '" title="' . $obj->name . '" ';
		$image .= 'width="' . $w . '" height="' . $h . '" />';
	}
	if ($obj->file =~ /\.mp[34]$/ or $obj->file =~ /\.m4a$/ or $obj->file =~ /\.pdf$/)
	{
		$option = '<!--[podcast]' . $obj->id . '[podcast]-->' . "\n";
	}
	my $path = $obj->get_url;
	my $player = $self->_get_player($path);
	$temp =~ s/%Image%/$image/g;
	$temp =~ s/%FilePath%/$path/g;
	$temp =~ s/%Player%/$player/g;
	return $temp . $option;
}

# ==================================================
package Sound::Info;
# --------------------------------------------------
# // Sound::Info - wrapper of MP3::Info, MP4::Info
# --------------------------------------------------
sub new
{
	my $class = shift;
	my $file = shift;
	my $self = {
		'title'    => '',
		'duration' => '',
		'artwork'  => undef,
		'arttype'  => '', # jpg or png
	};
	no strict 'refs';
	foreach my $member ( keys(%{$self}) )
	{ # create getters for instance member
		next if ( defined(&{$class . '::' . $member}) );
		*{$class . '::' . $member} = sub {
			my $self = shift;
			$self->{$member};
		};
	}
	bless($self,$class);
	return $self->_initialize($file);
}
sub _initialize
{
	my $self = shift;
	my $file = shift;
	if ($file =~ /\.mp3$/)
	{
		my $tags = &MP3::Info::get_mp3tag($file,2,2); # FILE, VERSION, RAW_V2
		my $info = &MP3::Info::get_mp3info($file);
		$self->{'artwork'} = &_extract_image($tags->{'PIC'}) if ($tags->{'PIC'});
		$self->{'title'} = ($tags->{'TT2'}) ? $tags->{'TT2'} : $tags->{'TIT2'};
		$self->{'duration'} = $info->{'TIME'};
	}
	elsif ($file =~ /\.m4a$|\.mp4$/)
	{
		my $mp4 = MP4::Info->new($file);
		$self->{'artwork'} = $mp4->covr;
		$self->{'title'} = $mp4->title;
		$self->{'duration'} = $mp4->time;
	}
	$self->_check_arttype if ($self->{'artwork'});
	return $self; # need to pass $self instance
}
sub _extract_image
{
	my $data = shift;
	# my $format = substr($data,0,3); # JPG or PNG
	# my $type = substr($data,3,1); # picture type
	$data = substr($data,5); # remove header
	$data;
	# [reference] format of PIC and APIC in ID3tag
	# http://silhouette.studio.mu/artwork.shtml
}
sub _check_arttype
{
	my $self = shift;
	my $pict = $self->artwork;
	my $head = substr($pict,0,10);
	$self->{'arttype'} = 
		  ($head eq "\x89\x50\x4e\x47\x0d\x0a\x1a\x0a\x00\x00") ? 'png'
		: ($head eq "\xff\xd8\xff\xe0\x00\x10\x4a\x46\x49\x46") ? 'jpg'
		: ''; # except png and jpg
}
# ==================================================
# Copyright (c) 1998-2005 Chris Nandor.  All rights reserved.
package MP3::Info;
require 5.006;
use overload;
use strict;
use Carp;
use vars qw(
	@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION $REVISION
	@mp3_genres %mp3_genres @winamp_genres %winamp_genres $try_harder
	@t_bitrate @t_sampling_freq @frequency_tbl %v1_tag_fields
	@v1_tag_names %v2_tag_names %v2_to_v1_names $AUTOLOAD
	@mp3_info_fields
);
@ISA = 'Exporter';
@EXPORT = qw(
	set_mp3tag get_mp3tag get_mp3info remove_mp3tag
	use_winamp_genres
);
@EXPORT_OK = qw(@mp3_genres %mp3_genres use_mp3_utf8);
%EXPORT_TAGS = (
	genres	=> [qw(@mp3_genres %mp3_genres)],
	utf8	=> [qw(use_mp3_utf8)],
	all	=> [@EXPORT, @EXPORT_OK]
);
# $Id: Info.pm,v 1.19 2005/03/11 04:41:29 pudge Exp $
($REVISION) = ' $Revision: 1.19 $ ' =~ /\$Revision:\s+([^\s]+)/;
$VERSION = '1.13';
sub new {
	my($pack, $file) = @_;
	my $info = get_mp3info($file) or return undef;
	my $tags = get_mp3tag($file) || { map { ($_ => undef) } @v1_tag_names };
	my %self = (
		FILE		=> $file,
		TRY_HARDER	=> 0
	);
	@self{@mp3_info_fields, @v1_tag_names, 'file'} = (
		@{$info}{@mp3_info_fields},
		@{$tags}{@v1_tag_names},
		$file
	);
	return bless \%self, $pack;
}
sub can {
	my $self = shift;
	return $self->SUPER::can(@_) unless ref $self;
	my $name = uc shift;
	return sub { $self->$name(@_) } if exists $self->{$name};
	return undef;
}
sub AUTOLOAD {
	my($self) = @_;
	(my $name = uc $AUTOLOAD) =~ s/^.*://;
	if (exists $self->{$name}) {
		my $sub = exists $v1_tag_fields{$name}
			? sub {
				if (defined $_[1]) {
					$_[0]->{$name} = $_[1];
					set_mp3tag($_[0]->{FILE}, $_[0]);
				}
				return $_[0]->{$name};
			}
			: sub {
				return $_[0]->{$name}
			};
		no strict 'refs';
		*{$AUTOLOAD} = $sub;
		goto &$AUTOLOAD;
	} else {
		carp(sprintf "No method '$name' available in package %s.",
			__PACKAGE__);
	}
}
sub DESTROY {
}
my $unicode_module = eval { require Encode; require Encode::Guess };
my $UNICODE = use_mp3_utf8($unicode_module ? 1 : 0);
sub use_mp3_utf8 {
	my($val) = @_;
	if ($val == 1) {
		if ($unicode_module) {
			$UNICODE = 1;
			$Encode::Guess::NoUTFAutoGuess = 1;
		}
	} elsif ($val == 0) {
		$UNICODE = 0;
	}
	return $UNICODE;
}
sub use_winamp_genres {
	%mp3_genres = %winamp_genres;
	@mp3_genres = @winamp_genres;
	return 1;
}
sub remove_mp3tag {
	my($file, $version, $buf) = @_;
	my($fh, $return);
	$buf ||= 4096*1024;  # the bigger the faster
	$version ||= 1;
	if (not (defined $file && $file ne '')) {
		$@ = "No file specified";
		return undef;
	}
	if (not -s $file) {
		$@ = "File is empty";
		return undef;
	}
	if (ref $file) { # filehandle passed
		$fh = $file;
	} else {
		if (not open $fh, '+<', $file) {
			$@ = "Can't open $file: $!";
			return undef;
		}
	}
	binmode $fh;
	if ($version eq 1 || $version eq 'ALL') {
		seek $fh, -128, 2;
		my $tell = tell $fh;
		if (<$fh> =~ /^TAG/) {
			truncate $fh, $tell or carp "Can't truncate '$file': $!";
			$return += 128;
		}
	}
	if ($version eq 2 || $version eq 'ALL') {
		my $v2h = _get_v2head($fh);
		if ($v2h) {
			local $\;
			seek $fh, 0, 2;
			my $eof = tell $fh;
			my $off = $v2h->{tag_size};
			while ($off < $eof) {
				seek $fh, $off, 0;
				read $fh, my($bytes), $buf;
				seek $fh, $off - $v2h->{tag_size}, 0;
				print $fh $bytes;
				$off += $buf;
			}
			truncate $fh, $eof - $v2h->{tag_size}
				or carp "Can't truncate '$file': $!";
			$return += $v2h->{tag_size};
		}
	}
	_close($file, $fh);
	return $return || -1;
}
sub set_mp3tag {
	my($file, $title, $artist, $album, $year, $comment, $genre, $tracknum) = @_;
	my(%info, $oldfh, $ref, $fh);
	local %v1_tag_fields = %v1_tag_fields;
	# set each to '' if undef
	for ($title, $artist, $album, $year, $comment, $tracknum, $genre,
		(@info{@v1_tag_names}))
		{$_ = defined() ? $_ : ''}
	($ref) = (overload::StrVal($title) =~ /^(?:.*\=)?([^=]*)\((?:[^\(]*)\)$/)
		if ref $title;
	# populate data to hashref if hashref is not passed
	if (!$ref) {
		(@info{@v1_tag_names}) =
			($title, $artist, $album, $year, $comment, $tracknum, $genre);
	# put data from hashref into hashref if hashref is passed
	} elsif ($ref eq 'HASH') {
		%info = %$title;
	# return otherwise
	} else {
		carp(<<'EOT');
Usage: set_mp3tag (FILE, TITLE, ARTIST, ALBUM, YEAR, COMMENT, GENRE [, TRACKNUM])
       set_mp3tag (FILE, $HASHREF)
EOT
		return undef;
	}
	if (not (defined $file && $file ne '')) {
		$@ = "No file specified";
		return undef;
	}
	if (not -s $file) {
		$@ = "File is empty";
		return undef;
	}
	# comment field length 28 if ID3v1.1
	$v1_tag_fields{COMMENT} = 28 if $info{TRACKNUM};

	# only if -w is on
	if ($^W) {
		# warn if fields too long
		foreach my $field (keys %v1_tag_fields) {
			$info{$field} = '' unless defined $info{$field};
			if (length($info{$field}) > $v1_tag_fields{$field}) {
				carp "Data too long for field $field: truncated to " .
					 "$v1_tag_fields{$field}";
			}
		}
		if ($info{GENRE}) {
			carp "Genre `$info{GENRE}' does not exist\n"
				unless exists $mp3_genres{$info{GENRE}};
		}
	}
	if ($info{TRACKNUM}) {
		$info{TRACKNUM} =~ s/^(\d+)\/(\d+)$/$1/;
		unless ($info{TRACKNUM} =~ /^\d+$/ &&
			$info{TRACKNUM} > 0 && $info{TRACKNUM} < 256) {
			carp "Tracknum `$info{TRACKNUM}' must be an integer " .
				"from 1 and 255\n" if $^W;
			$info{TRACKNUM} = '';
		}
	}
	if (ref $file) { # filehandle passed
		$fh = $file;
	} else {
		if (not open $fh, '+<', $file) {
			$@ = "Can't open $file: $!";
			return undef;
		}
	}
	binmode $fh;
	$oldfh = select $fh;
	seek $fh, -128, 2;
	# go to end of file if no tag, beginning of file if tag
	seek $fh, (<$fh> =~ /^TAG/ ? -128 : 0), 2;
	# get genre value
	$info{GENRE} = $info{GENRE} && exists $mp3_genres{$info{GENRE}} ?
		$mp3_genres{$info{GENRE}} : 255;  # some default genre
	local $\;
	# print TAG to file
	if ($info{TRACKNUM}) {
		print pack 'a3a30a30a30a4a28xCC', 'TAG', @info{@v1_tag_names};
	} else {
		print pack 'a3a30a30a30a4a30C', 'TAG', @info{@v1_tag_names[0..4, 6]};
	}
	select $oldfh;
	_close($file, $fh);
	return 1;
}
sub get_mp3tag {
	my($file, $ver, $raw_v2) = @_;
	my($tag, $v1, $v2, $v2h, %info, @array, $fh);
	$raw_v2 ||= 0;
	$ver = !$ver ? 0 : ($ver == 2 || $ver == 1) ? $ver : 0;
	if (not (defined $file && $file ne '')) {
		$@ = "No file specified";
		return undef;
	}
	if (not -s $file) {
		$@ = "File is empty";
		return undef;
	}
	if (ref $file) { # filehandle passed
		$fh = $file;
	} else {
		if (not open $fh, '<', $file) {
			$@ = "Can't open $file: $!";
			return undef;
		}
	}
	binmode $fh;
	if ($ver < 2) {
		seek $fh, -128, 2;
		while(defined(my $line = <$fh>)) { $tag .= $line }
		if ($tag && $tag =~ /^TAG/) {
			$v1 = 1;
			if (substr($tag, -3, 2) =~ /\000[^\000]/) {
				(undef, @info{@v1_tag_names}) =
					(unpack('a3a30a30a30a4a28', $tag),
					ord(substr($tag, -2, 1)),
					$mp3_genres[ord(substr $tag, -1)]);
				$info{TAGVERSION} = 'ID3v1.1';
			} else {
				(undef, @info{@v1_tag_names[0..4, 6]}) =
					(unpack('a3a30a30a30a4a30', $tag),
					$mp3_genres[ord(substr $tag, -1)]);
				$info{TAGVERSION} = 'ID3v1';
			}
			if ($UNICODE) {
				for my $key (keys %info) {
					next unless $info{$key};
					$info{$key} = Encode::encode_utf8($info{$key});
				}
			}
		} elsif ($ver == 1) {
			_close($file, $fh);
			$@ = "No ID3v1 tag found";
			return undef;
		}
	}
	($v2, $v2h) = _get_v2tag($fh);
	unless ($v1 || $v2) {
		_close($file, $fh);
		$@ = "No ID3 tag found";
		return undef;
	}
	if (($ver == 0 || $ver == 2) && $v2) {
		if ($raw_v2 == 1 && $ver == 2) {
			%info = %$v2;
			$info{TAGVERSION} = $v2h->{version};
		} else {
			my $hash = $raw_v2 == 2 ? { map { ($_, $_) } keys %v2_tag_names } : \%v2_to_v1_names;
			for my $id (keys %$hash) {
				if (exists $v2->{$id}) {
					my $data1 = $v2->{$id};
					# this is tricky ... if this is an arrayref,
					# we want to only return one, so we pick the
					# first one.  but if it is a comment, we pick
					# the first one where the first charcter after
					# the language is NULL and not an additional
					# sub-comment, because that is most likely to be
					# the user-supplied comment
					if (ref $data1 && !$raw_v2) {
						if ($id =~ /^COMM?$/) {
							my($newdata) = grep /^(....\000)/, @{$data1};
							$data1 = $newdata || $data1->[0];
						} else {
							$data1 = $data1->[0];
						}
					}
					$data1 = [ $data1 ] if ! ref $data1;
					for my $data (@$data1) {
						# TODO : this should only be done for certain frames;
						# using RAW still gives you access, but we should be smarter
						# about how individual frame types are handled.  it's not
						# like the list is infinitely long.
						$data =~ s/^(.)//; # strip first char (text encoding)
						my $encoding = $1;
						my $desc;
						if ($id =~ /^COM[M ]?$/) { # space for iTunes brokenness
							$data =~ s/^(?:...)//;		# strip language
						}
						if ($UNICODE) {
							if ($encoding eq "\001" || $encoding eq "\002") {  # UTF-16, UTF-16BE
								# text fields can be null-separated lists;
								# UTF-16 therefore needs special care
								#
								# foobar2000 encodes tags in UTF-16LE
								# (which is apparently illegal)
								# Encode dies on a bad BOM, so it is
								# probably wise to wrap it in an eval
								# anyway
								$data = eval { Encode::decode('utf16', $data) } || Encode::decode('utf16le', $data);
								# this split we do doesn't work, because obviously
								# two NULLs can appear where we don't want ...
								#$data = join "\000", map {
								#	eval { Encode::decode('utf16', $_) } || Encode::decode('utf16le', $_)
								#} split /\000\000/, $data;
							} elsif ($encoding eq "\003") { # UTF-8
								# make sure string is UTF8, and set flag appropriately
								$data = Encode::decode('utf8', $data);
							} elsif ($encoding eq "\000") {
								# Try and guess the encoding, otherwise just use latin1
								my $dec = Encode::Guess->guess($data);
								if (ref $dec) {
									$data = $dec->decode($data);
								} else {
									# Best try
									$data = Encode::decode('iso-8859-1', $data);
								}
							}
							# do we care about trailing NULL?
							# $data =~ s/\000$//;
						} else {
							# If the string starts with an
							# UTF-16 little endian BOM, use a hack to
							# convert to ASCII per best-effort
							my $pat;
							if ($data =~ s/^\xFF\xFE//) {
								$pat = 'v';
							} elsif ($data =~ s/^\xFE\xFF//) {
								$pat = 'n';
							}
							if ($pat) {
								$data = pack 'C*', map {
									(chr =~ /[[:ascii:]]/ && chr =~ /[[:print:]]/)
										? $_
										: ord('?')
								} unpack "$pat*", $data;
							}
						}
						# We do this after decoding so we could be certain we're dealing
						# with 8-bit text.
						if ($id =~ /^COM[M ]?$/) { # space for iTunes brokenness
							$data =~ s/^(.*?)\000//;	# strip up to first NULL(s),
											# for sub-comments (TODO:
											# handle all comment data)
							$desc = $1;
						} elsif ($id =~ /^TCON?$/) {
							if ($data =~ /^ \(? (\d+) (?:\)|\000)? (.+)?/sx) {
								my($index, $name) = ($1, $2);
								if ($name && $name ne "\000") {
									$data = $name;
								} else {
									$data = $mp3_genres[$index];
								}
							}
						}
						if ($raw_v2 == 2 && $desc) {
							$data = { $desc => $data };
						}
						if ($raw_v2 == 2 && exists $info{$hash->{$id}}) {
							if (ref $info{$hash->{$id}} eq 'ARRAY') {
								push @{$info{$hash->{$id}}}, $data;
							} else {
								$info{$hash->{$id}} = [ $info{$hash->{$id}}, $data ];
							}
						} else {
							$info{$hash->{$id}} = $data;
						}
					}
				}
			}
			if ($ver == 0 && $info{TAGVERSION}) {
				$info{TAGVERSION} .= ' / ' . $v2h->{version};
			} else {
				$info{TAGVERSION} = $v2h->{version};
			}
		}
	}
	unless ($raw_v2 && $ver == 2) {
		foreach my $key (keys %info) {
			if (defined $info{$key}) {
				$info{$key} =~ s/\000+.*//g;
				$info{$key} =~ s/\s+$//;
			}
		}
		for (@v1_tag_names) {
			$info{$_} = '' unless defined $info{$_};
		}
	}
	if (keys %info && exists $info{GENRE} && ! defined $info{GENRE}) {
		$info{GENRE} = '';
	}
	_close($file, $fh);
	return keys %info ? {%info} : undef;
}
sub _get_v2tag {
	my($fh) = @_;
	my($off, $end, $myseek, $v2, $v2h, $hlen, $num, $wholetag);
	$v2 = {};
	$v2h = _get_v2head($fh) or return;
	if ($v2h->{major_version} < 2) {
		carp "This is $v2h->{version}; " .
		     "ID3v2 versions older than ID3v2.2.0 not supported\n"
		     if $^W;
		return;
	}
	# use syncsafe bytes if using version 2.4
	my $bytesize = ($v2h->{major_version} > 3) ? 128 : 256;
	if ($v2h->{major_version} == 2) {
		$hlen = 6;
		$num = 3;
	} else {
		$hlen = 10;
		$num = 4;
	}
	$off = $v2h->{ext_header_size} + 10;
	$end = $v2h->{tag_size} + 10; # should we read in the footer too?
	seek $fh, $v2h->{offset}, 0;
	read $fh, $wholetag, $end;
	$wholetag =~ s/\xFF\x00/\xFF/gs if $v2h->{unsync};
	$myseek = sub {
		my $bytes = substr($wholetag, $off, $hlen);
		return unless $bytes =~ /^([A-Z0-9]{$num})/
			|| ($num == 4 && $bytes =~ /^(COM )/);  # stupid iTunes
		my($id, $size) = ($1, $hlen);
		my @bytes = reverse unpack "C$num", substr($bytes, $num, $num);
		for my $i (0 .. ($num - 1)) {
			$size += $bytes[$i] * $bytesize ** $i;
		}
		my $flags = {};
		if ($v2h->{major_version} > 3) {
			my @bits = split //, unpack 'B16', substr($bytes, 8, 2);
			$flags->{frame_unsync}       = $bits[14];
			$flags->{data_len_indicator} = $bits[15];
		}
		return($id, $size, $flags);
	};
	while ($off < $end) {
		my($id, $size, $flags) = &$myseek or last;
		my $bytes = substr($wholetag, $off+$hlen, $size-$hlen);
		my $data_len;
		if ($flags->{data_len_indicator}) {
			$data_len = 0;
			my @data_len_bytes = reverse unpack 'C4', substr($bytes, 0, 4);
			$bytes = substr($bytes, 4);
		        for my $i (0..3) {
				$data_len += $data_len_bytes[$i] * 128 ** $i;
		        }
		}
		# perform frame-level unsync if needed (skip if already done for whole tag)
		$bytes =~ s/\xFF\x00/\xFF/gs if $flags->{frame_unsync} && !$v2h->{unsync};
		# if we know the data length, sanity check it now.
		if ($flags->{data_len_indicator} && defined $data_len) {
		        carp "Size mismatch on $id\n" unless $data_len == length($bytes);
		}
		if (exists $v2->{$id}) {
			if (ref $v2->{$id} eq 'ARRAY') {
				push @{$v2->{$id}}, $bytes;
			} else {
				$v2->{$id} = [$v2->{$id}, $bytes];
			}
		} else {
			$v2->{$id} = $bytes;
		}
		$off += $size;
	}
	return($v2, $v2h);
}
sub get_mp3info {
	my($file) = @_;
	my($off, $byte, $eof, $h, $tot, $fh);
	if (not (defined $file && $file ne '')) {
		$@ = "No file specified";
		return undef;
	}
	if (not -s $file) {
		$@ = "File is empty";
		return undef;
	}
	if (ref $file) { # filehandle passed
		$fh = $file;
	} else {
		if (not open $fh, '<', $file) {
			$@ = "Can't open $file: $!";
			return undef;
		}
	}
	$off = 0;
	$tot = 8192;
	binmode $fh;
	seek $fh, $off, 0;
	read $fh, $byte, 4;
	if ($off == 0) {
		if (my $v2h = _get_v2head($fh)) {
			$tot += $off += $v2h->{tag_size};
			seek $fh, $off, 0;
			read $fh, $byte, 4;
		}
	}
	$h = _get_head($byte);
	my $is_mp3 = _is_mp3($h); 
	until ($is_mp3) {
		$off++;
		seek $fh, $off, 0;
		read $fh, $byte, 4;
		if ($off > $tot && !$try_harder) {
			_close($file, $fh);
			$@ = "Couldn't find MP3 header (perhaps set " .
			     '$MP3::Info::try_harder and retry)';
			return undef;
		}
		next if ord($byte) != 0xFF;
		$h = _get_head($byte);
		$is_mp3 = _is_mp3($h);
	}
	my $vbr = _get_vbr($fh, $h, \$off);
	seek $fh, 0, 2;
	$eof = tell $fh;
	seek $fh, -128, 2;
	$eof -= 128 if <$fh> =~ /^TAG/ ? 1 : 0;
	_close($file, $fh);
	$h->{size} = $eof - $off;
	$h->{offset} = $off;
	return _get_info($h, $vbr);
}
sub _get_info {
	my($h, $vbr) = @_;
	my $i;
	$i->{VERSION}	= $h->{IDR} == 2 ? 2 : $h->{IDR} == 3 ? 1 :
				$h->{IDR} == 0 ? 2.5 : 0;
	$i->{LAYER}	= 4 - $h->{layer};
	$i->{VBR}	= defined $vbr ? 1 : 0;
	$i->{COPYRIGHT}	= $h->{copyright} ? 1 : 0;
	$i->{PADDING}	= $h->{padding_bit} ? 1 : 0;
	$i->{STEREO}	= $h->{mode} == 3 ? 0 : 1;
	$i->{MODE}	= $h->{mode};
	$i->{SIZE}	= $vbr && $vbr->{bytes} ? $vbr->{bytes} : $h->{size};
	$i->{OFFSET}	= $h->{offset};
	my $mfs		= $h->{fs} / ($h->{ID} ? 144000 : 72000);
	$i->{FRAMES}	= int($vbr && $vbr->{frames}
				? $vbr->{frames}
				: $i->{SIZE} / ($h->{bitrate} / $mfs)
			  );
	if ($vbr) {
		$i->{VBR_SCALE}	= $vbr->{scale} if $vbr->{scale};
		$h->{bitrate}	= $i->{SIZE} / $i->{FRAMES} * $mfs;
		if (not $h->{bitrate}) {
			$@ = "Couldn't determine VBR bitrate";
			return undef;
		}
	}
	$h->{'length'}	= ($i->{SIZE} * 8) / $h->{bitrate} / 10;
	$i->{SECS}	= $h->{'length'} / 100;
	$i->{MM}	= int $i->{SECS} / 60;
	$i->{SS}	= int $i->{SECS} % 60;
	$i->{MS}	= (($i->{SECS} - ($i->{MM} * 60) - $i->{SS}) * 1000);
#	$i->{LF}	= ($i->{MS} / 1000) * ($i->{FRAMES} / $i->{SECS});
#	int($i->{MS} / 100 * 75);  # is this right?
	$i->{TIME}	= sprintf "%.2d:%.2d", @{$i}{'MM', 'SS'};
	$i->{BITRATE}		= int $h->{bitrate};
	# should we just return if ! FRAMES?
	$i->{FRAME_LENGTH}	= int($h->{size} / $i->{FRAMES}) if $i->{FRAMES};
	$i->{FREQUENCY}		= $frequency_tbl[3 * $h->{IDR} + $h->{sampling_freq}];
	return $i;
}
sub _get_head {
	my($byte) = @_;
	my($bytes, $h);
	$bytes = _unpack_head($byte);
	@$h{qw(IDR ID layer protection_bit
		bitrate_index sampling_freq padding_bit private_bit
		mode mode_extension copyright original
		emphasis version_index bytes)} = (
		($bytes>>19)&3, ($bytes>>19)&1, ($bytes>>17)&3, ($bytes>>16)&1,
		($bytes>>12)&15, ($bytes>>10)&3, ($bytes>>9)&1, ($bytes>>8)&1,
		($bytes>>6)&3, ($bytes>>4)&3, ($bytes>>3)&1, ($bytes>>2)&1,
		$bytes&3, ($bytes>>19)&3, $bytes
	);
	$h->{bitrate} = $t_bitrate[$h->{ID}][3 - $h->{layer}][$h->{bitrate_index}];
	$h->{fs} = $t_sampling_freq[$h->{IDR}][$h->{sampling_freq}];
	return $h;
}
sub _is_mp3 {
	my $h = $_[0] or return undef;
	return ! (	# all below must be false
		 $h->{bitrate_index} == 0
			||
		 $h->{version_index} == 1
			||
		($h->{bytes} & 0xFFE00000) != 0xFFE00000
			||
		!$h->{fs}
			||
		!$h->{bitrate}
			||
		 $h->{bitrate_index} == 15
			||
		!$h->{layer}
			||
		 $h->{sampling_freq} == 3
			||
		 $h->{emphasis} == 2
			||
		!$h->{bitrate_index}
			||
		($h->{bytes} & 0xFFFF0000) == 0xFFFE0000
			||
		($h->{ID} == 1 && $h->{layer} == 3 && $h->{protection_bit} == 1)
			||
		($h->{mode_extension} != 0 && $h->{mode} != 1)
	);
}
sub _get_vbr {
	my($fh, $h, $roff) = @_;
	my($off, $bytes, @bytes, $myseek, %vbr);
	$off = $$roff;
	@_ = ();	# closure confused if we don't do this
	$myseek = sub {
		my $n = $_[0] || 4;
		seek $fh, $off, 0;
		read $fh, $bytes, $n;
		$off += $n;
	};
	$off += 4;
	if ($h->{ID}) {	# MPEG1
		$off += $h->{mode} == 3 ? 17 : 32;
	} else {	# MPEG2
		$off += $h->{mode} == 3 ? 9 : 17;
	}
	&$myseek;
	return unless $bytes eq 'Xing';
	&$myseek;
	$vbr{flags} = _unpack_head($bytes);
	if ($vbr{flags} & 1) {
		&$myseek;
		$vbr{frames} = _unpack_head($bytes);
	}
	if ($vbr{flags} & 2) {
		&$myseek;
		$vbr{bytes} = _unpack_head($bytes);
	}
	if ($vbr{flags} & 4) {
		$myseek->(100);
# Not used right now ...
#		$vbr{toc} = _unpack_head($bytes);
	}
	if ($vbr{flags} & 8) { # (quality ind., 0=best 100=worst)
		&$myseek;
		$vbr{scale} = _unpack_head($bytes);
	} else {
		$vbr{scale} = -1;
	}
	$$roff = $off;
	return \%vbr;
}
sub _get_v2head {
	my $fh = $_[0] or return;
	my($v2h, $bytes, @bytes);
	$v2h->{offset} = 0;
	# check first three bytes for 'ID3'
	seek $fh, 0, 0;
	read $fh, $bytes, 3;
	# TODO: add support for tags at the end of the file
	if ($bytes eq 'RIF' || $bytes eq 'FOR') {
		_find_id3_chunk($fh, $bytes) or return;
		$v2h->{offset} = tell $fh;
		read $fh, $bytes, 3;
	}
	return unless $bytes eq 'ID3';
	# get version
	read $fh, $bytes, 2;
	$v2h->{version} = sprintf "ID3v2.%d.%d",
		@$v2h{qw[major_version minor_version]} =
			unpack 'c2', $bytes;
	# get flags
	read $fh, $bytes, 1;
	my @bits = split //, unpack 'b8', $bytes;
	if ($v2h->{major_version} == 2) {
		$v2h->{unsync}       = $bits[7];
		$v2h->{compression}  = $bits[8];
		$v2h->{ext_header}   = 0;
		$v2h->{experimental} = 0;
	} else {
		$v2h->{unsync}       = $bits[7];
		$v2h->{ext_header}   = $bits[6];
		$v2h->{experimental} = $bits[5];
		$v2h->{footer}       = $bits[4] if $v2h->{major_version} == 4;
	}
	# get ID3v2 tag length from bytes 7-10
	$v2h->{tag_size} = 10;	# include ID3v2 header size
	$v2h->{tag_size} += 10 if $v2h->{footer};
	read $fh, $bytes, 4;
	@bytes = reverse unpack 'C4', $bytes;
	foreach my $i (0 .. 3) {
		# whoaaaaaa nellllllyyyyyy!
		$v2h->{tag_size} += $bytes[$i] * 128 ** $i;
	}
	# get extended header size
	$v2h->{ext_header_size} = 0;
	if ($v2h->{ext_header}) {
		read $fh, $bytes, 4;
		@bytes = reverse unpack 'C4', $bytes;
		# use syncsafe bytes if using version 2.4
		my $bytesize = ($v2h->{major_version} > 3) ? 128 : 256;
		for my $i (0..3) {
			$v2h->{ext_header_size} += $bytes[$i] * $bytesize ** $i;
		}
	}
	return $v2h;
}
sub _find_id3_chunk {
	my($fh, $filetype) = @_;
	my($bytes, $size, $tag, $pat, $mat);
	read $fh, $bytes, 1;
	if ($filetype eq 'RIF') {  # WAV
		return 0 if $bytes ne 'F';
		$pat = 'a4V';
		$mat = 'id3 ';
	} elsif ($filetype eq 'FOR') { # AIFF
		return 0 if $bytes ne 'M';
		$pat = 'a4N';
		$mat = 'ID3 ';
	}
	seek $fh, 12, 0;  # skip to the first chunk
	while ((read $fh, $bytes, 8) == 8) {
		($tag, $size)  = unpack $pat, $bytes;
		return 1 if $tag eq $mat;
		seek $fh, $size, 1;
	}
	return 0;
}
sub _unpack_head {
	unpack('l', pack('L', unpack('N', $_[0])));
}
sub _close {
	my($file, $fh) = @_;
	unless (ref $file) { # filehandle not passed
		close $fh or carp "Problem closing '$file': $!";
	}
}
BEGIN {
	@mp3_genres = (
		'Blues',
		'Classic Rock',
		'Country',
		'Dance',
		'Disco',
		'Funk',
		'Grunge',
		'Hip-Hop',
		'Jazz',
		'Metal',
		'New Age',
		'Oldies',
		'Other',
		'Pop',
		'R&B',
		'Rap',
		'Reggae',
		'Rock',
		'Techno',
		'Industrial',
		'Alternative',
		'Ska',
		'Death Metal',
		'Pranks',
		'Soundtrack',
		'Euro-Techno',
		'Ambient',
		'Trip-Hop',
		'Vocal',
		'Jazz+Funk',
		'Fusion',
		'Trance',
		'Classical',
		'Instrumental',
		'Acid',
		'House',
		'Game',
		'Sound Clip',
		'Gospel',
		'Noise',
		'AlternRock',
		'Bass',
		'Soul',
		'Punk',
		'Space',
		'Meditative',
		'Instrumental Pop',
		'Instrumental Rock',
		'Ethnic',
		'Gothic',
		'Darkwave',
		'Techno-Industrial',
		'Electronic',
		'Pop-Folk',
		'Eurodance',
		'Dream',
		'Southern Rock',
		'Comedy',
		'Cult',
		'Gangsta',
		'Top 40',
		'Christian Rap',
		'Pop/Funk',
		'Jungle',
		'Native American',
		'Cabaret',
		'New Wave',
		'Psychadelic',
		'Rave',
		'Showtunes',
		'Trailer',
		'Lo-Fi',
		'Tribal',
		'Acid Punk',
		'Acid Jazz',
		'Polka',
		'Retro',
		'Musical',
		'Rock & Roll',
		'Hard Rock',
	);
	@winamp_genres = (@mp3_genres);
	@t_bitrate = ([
		[0, 32, 48, 56,  64,  80,  96, 112, 128, 144, 160, 176, 192, 224, 256],
		[0,  8, 16, 24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160],
		[0,  8, 16, 24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160]
	],[
		[0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448],
		[0, 32, 48, 56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384],
		[0, 32, 40, 48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320]
	]);
	@t_sampling_freq = (
		[11025, 12000,  8000],
		[undef, undef, undef],	# reserved
		[22050, 24000, 16000],
		[44100, 48000, 32000]
	);
	@frequency_tbl = map { $_ ? eval "${_}e-3" : 0 }
		map { @$_ } @t_sampling_freq;
	@mp3_info_fields = qw(
		VERSION
		LAYER
		STEREO
		VBR
		BITRATE
		FREQUENCY
		SIZE
		OFFSET
		SECS
		MM
		SS
		MS
		TIME
		COPYRIGHT
		PADDING
		MODE
		FRAMES
		FRAME_LENGTH
		VBR_SCALE
	);
	%v1_tag_fields = (TITLE => 30, ARTIST => 30, ALBUM => 30, COMMENT => 30, YEAR => 4);
	@v1_tag_names = qw(TITLE ARTIST ALBUM YEAR COMMENT TRACKNUM GENRE);
	%v2_to_v1_names = (
		# v2.2 tags
		'TT2' => 'TITLE',
		'TP1' => 'ARTIST',
		'TAL' => 'ALBUM',
		'TYE' => 'YEAR',
		'COM' => 'COMMENT',
		'TRK' => 'TRACKNUM',
		'TCO' => 'GENRE', # not clean mapping, but ...
		# v2.3 tags
		'TIT2' => 'TITLE',
		'TPE1' => 'ARTIST',
		'TALB' => 'ALBUM',
		'TYER' => 'YEAR',
		'COMM' => 'COMMENT',
		'TRCK' => 'TRACKNUM',
		'TCON' => 'GENRE',
	);
	%v2_tag_names = (
		# v2.2 tags
		'BUF' => 'Recommended buffer size',
		'CNT' => 'Play counter',
		'COM' => 'Comments',
		'CRA' => 'Audio encryption',
		'CRM' => 'Encrypted meta frame',
		'ETC' => 'Event timing codes',
		'EQU' => 'Equalization',
		'GEO' => 'General encapsulated object',
		'IPL' => 'Involved people list',
		'LNK' => 'Linked information',
		'MCI' => 'Music CD Identifier',
		'MLL' => 'MPEG location lookup table',
		'PIC' => 'Attached picture',
		'POP' => 'Popularimeter',
		'REV' => 'Reverb',
		'RVA' => 'Relative volume adjustment',
		'SLT' => 'Synchronized lyric/text',
		'STC' => 'Synced tempo codes',
		'TAL' => 'Album/Movie/Show title',
		'TBP' => 'BPM (Beats Per Minute)',
		'TCM' => 'Composer',
		'TCO' => 'Content type',
		'TCR' => 'Copyright message',
		'TDA' => 'Date',
		'TDY' => 'Playlist delay',
		'TEN' => 'Encoded by',
		'TFT' => 'File type',
		'TIM' => 'Time',
		'TKE' => 'Initial key',
		'TLA' => 'Language(s)',
		'TLE' => 'Length',
		'TMT' => 'Media type',
		'TOA' => 'Original artist(s)/performer(s)',
		'TOF' => 'Original filename',
		'TOL' => 'Original Lyricist(s)/text writer(s)',
		'TOR' => 'Original release year',
		'TOT' => 'Original album/Movie/Show title',
		'TP1' => 'Lead artist(s)/Lead performer(s)/Soloist(s)/Performing group',
		'TP2' => 'Band/Orchestra/Accompaniment',
		'TP3' => 'Conductor/Performer refinement',
		'TP4' => 'Interpreted, remixed, or otherwise modified by',
		'TPA' => 'Part of a set',
		'TPB' => 'Publisher',
		'TRC' => 'ISRC (International Standard Recording Code)',
		'TRD' => 'Recording dates',
		'TRK' => 'Track number/Position in set',
		'TSI' => 'Size',
		'TSS' => 'Software/hardware and settings used for encoding',
		'TT1' => 'Content group description',
		'TT2' => 'Title/Songname/Content description',
		'TT3' => 'Subtitle/Description refinement',
		'TXT' => 'Lyricist/text writer',
		'TXX' => 'User defined text information frame',
		'TYE' => 'Year',
		'UFI' => 'Unique file identifier',
		'ULT' => 'Unsychronized lyric/text transcription',
		'WAF' => 'Official audio file webpage',
		'WAR' => 'Official artist/performer webpage',
		'WAS' => 'Official audio source webpage',
		'WCM' => 'Commercial information',
		'WCP' => 'Copyright/Legal information',
		'WPB' => 'Publishers official webpage',
		'WXX' => 'User defined URL link frame',
		# v2.3 tags
		'AENC' => 'Audio encryption',
		'APIC' => 'Attached picture',
		'COMM' => 'Comments',
		'COMR' => 'Commercial frame',
		'ENCR' => 'Encryption method registration',
		'EQUA' => 'Equalization',
		'ETCO' => 'Event timing codes',
		'GEOB' => 'General encapsulated object',
		'GRID' => 'Group identification registration',
		'IPLS' => 'Involved people list',
		'LINK' => 'Linked information',
		'MCDI' => 'Music CD identifier',
		'MLLT' => 'MPEG location lookup table',
		'OWNE' => 'Ownership frame',
		'PCNT' => 'Play counter',
		'POPM' => 'Popularimeter',
		'POSS' => 'Position synchronisation frame',
		'PRIV' => 'Private frame',
		'RBUF' => 'Recommended buffer size',
		'RVAD' => 'Relative volume adjustment',
		'RVRB' => 'Reverb',
		'SYLT' => 'Synchronized lyric/text',
		'SYTC' => 'Synchronized tempo codes',
		'TALB' => 'Album/Movie/Show title',
		'TBPM' => 'BPM (beats per minute)',
		'TCOM' => 'Composer',
		'TCON' => 'Content type',
		'TCOP' => 'Copyright message',
		'TDAT' => 'Date',
		'TDLY' => 'Playlist delay',
		'TENC' => 'Encoded by',
		'TEXT' => 'Lyricist/Text writer',
		'TFLT' => 'File type',
		'TIME' => 'Time',
		'TIT1' => 'Content group description',
		'TIT2' => 'Title/songname/content description',
		'TIT3' => 'Subtitle/Description refinement',
		'TKEY' => 'Initial key',
		'TLAN' => 'Language(s)',
		'TLEN' => 'Length',
		'TMED' => 'Media type',
		'TOAL' => 'Original album/movie/show title',
		'TOFN' => 'Original filename',
		'TOLY' => 'Original lyricist(s)/text writer(s)',
		'TOPE' => 'Original artist(s)/performer(s)',
		'TORY' => 'Original release year',
		'TOWN' => 'File owner/licensee',
		'TPE1' => 'Lead performer(s)/Soloist(s)',
		'TPE2' => 'Band/orchestra/accompaniment',
		'TPE3' => 'Conductor/performer refinement',
		'TPE4' => 'Interpreted, remixed, or otherwise modified by',
		'TPOS' => 'Part of a set',
		'TPUB' => 'Publisher',
		'TRCK' => 'Track number/Position in set',
		'TRDA' => 'Recording dates',
		'TRSN' => 'Internet radio station name',
		'TRSO' => 'Internet radio station owner',
		'TSIZ' => 'Size',
		'TSRC' => 'ISRC (international standard recording code)',
		'TSSE' => 'Software/Hardware and settings used for encoding',
		'TXXX' => 'User defined text information frame',
		'TYER' => 'Year',
		'UFID' => 'Unique file identifier',
		'USER' => 'Terms of use',
		'USLT' => 'Unsychronized lyric/text transcription',
		'WCOM' => 'Commercial information',
		'WCOP' => 'Copyright/Legal information',
		'WOAF' => 'Official audio file webpage',
		'WOAR' => 'Official artist/performer webpage',
		'WOAS' => 'Official audio source webpage',
		'WORS' => 'Official internet radio station homepage',
		'WPAY' => 'Payment',
		'WPUB' => 'Publishers official webpage',
		'WXXX' => 'User defined URL link frame',
		# v2.4 additional tags
		# note that we don't restrict tags from 2.3 or 2.4,
		'ASPI' => 'Audio seek point index',
		'EQU2' => 'Equalisation (2)',
		'RVA2' => 'Relative volume adjustment (2)',
		'SEEK' => 'Seek frame',
		'SIGN' => 'Signature frame',
		'TDEN' => 'Encoding time',
		'TDOR' => 'Original release time',
		'TDRC' => 'Recording time',
		'TDRL' => 'Release time',
		'TDTG' => 'Tagging time',
		'TIPL' => 'Involved people list',
		'TMCL' => 'Musician credits list',
		'TMOO' => 'Mood',
		'TPRO' => 'Produced notice',
		'TSOA' => 'Album sort order',
		'TSOP' => 'Performer sort order',
		'TSOT' => 'Title sort order',
		'TSST' => 'Set subtitle',
		# grrrrrrr
		'COM ' => 'Broken iTunes comments',
	);
}
# ==================================================
# Copyright (c) 2004, 2005, Jonathan Harris <jhar@cpan.org>
package MP4::Info;
use overload;
use strict;
use Carp;
use Symbol;
my $enable_encoder = undef;
eval {
	use Encode;
	use Encode::Guess qw(latin1);
};
my $enable_encode = 1 if (!$@);
use vars qw(
	    $VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $AUTOLOAD
	    %data_atoms %other_atoms %container_atoms @mp4_genres
	   );
@ISA = 'Exporter';
@EXPORT      = qw(get_mp4tag get_mp4info);
@EXPORT_OK   = qw(use_mp4_utf8);
%EXPORT_TAGS = (
		utf8	=> [qw(use_mp4_utf8)],
		all	=> [@EXPORT, @EXPORT_OK]
	       );
$VERSION = '1.05';
my $debug = 0;
sub new
{
    my ($class, $file) = @_;
    # Supported tags
    my %tag_names =
	(
	 ALB => 1, APID => 1, ART => 1, CMT => 1, COVR => 1, CPIL => 1, CPRT => 1, DAY => 1, DISK => 1, GNRE => 1, GRP => 1, NAM => 1, RTNG => 1, TMPO => 1, TOO => 1, TRKN => 1, WRT => 1,
	 TITLE => 1, ARTIST => 1, ALBUM => 1, YEAR => 1, COMMENT => 1, GENRE => 1, TRACKNUM => 1,
	 VERSION => 1, LAYER => 1,
	 BITRATE => 1, FREQUENCY => 1, SIZE => 1,
	 SECS => 1, MM => 1, SS => 1, MS => 1, TIME => 1,
	 COPYRIGHT => 1, ENCRYPTED => 1,
	);
    my $tags = get_mp4tag ($file) or return undef;
    my $self = {
		_permitted => \%tag_names,
		%$tags
	       };
    return bless $self, $class;
}
sub AUTOLOAD
{
    my $self = shift;
    my $type = ref($self) or croak "$self is not an object";
    my $name = $AUTOLOAD;
    $name =~ s/.*://;	# strip fully-qualified portion
    unless (exists $self->{_permitted}->{uc $name} )
    {
	croak "No method '$name' available in class $type";
    }
    # Ignore any parameter
    return $self->{uc $name};
}
sub DESTROY
{
}
my $utf8 = 1;
sub use_mp4_utf8
{
    my ($val) = @_;
    $utf8 = $val if (($val == 0) || ($val == 1));
    return $utf8;
}
sub get_mp4tag
{
    my ($file) = @_;
    my (%tags);
    return parse_file ($file, \%tags) ? undef : {%tags};
}
sub get_mp4info
{
    my ($file) = @_;
    my (%tags);
    return parse_file ($file, \%tags) ? undef : {%tags};
}
my %data_atoms =
    (
     AART => 1,	# Album artist - returned in ART field no ART found
     ALB  => 1,
     ART  => 1,
     CMT  => 1,
     COVR => 1, # Cover art
     CPIL => 1,
     CPRT => 1,
     DAY  => 1,
     DISK => 1,
     GEN  => 1,	# Custom genre - returned in GNRE field no GNRE found
     GNRE => 1,	# Standard ID3/WinAmp genre
     GRP  => 1,
     NAM  => 1,
     RTNG => 1,
     TMPO => 1,
     TOO  => 1,
     TRKN => 1,
     WRT  => 1,
     # Apple store
     APID => 1,
     AKID => 1,	# ???
     ATID => 1,	# ???
     CNID => 1,	# ???
     GEID => 1,	# Some kind of watermarking ???
     PLID => 1,	# ???
     # 3GPP
     TITL => 1,	# title       - returned in NAM field no NAM found
     DSCP => 1, # description - returned in CMT field no CMT found
     #CPRT=> 1,
     PERF => 1, # performer   - returned in ART field no ART found
     AUTH => 1,	# author      - returned in WRT field no WRT found
     #GNRE=> 1,
    );
my %other_atoms =
    (
     MDAT => \&parse_mdat,
     META => \&parse_meta,
     MVHD => \&parse_mvhd,
     STSD => \&parse_stsd,
    );
my %container_atoms =
    (
     ILST => 1,
     MDIA => 1,
     MINF => 1,
     MOOV => 1,
     STBL => 1,
     TRAK => 1,
     UDTA => 1,
    );
my @mp4_genres =
    (
     'N/A', 'Blues', 'Classic Rock', 'Country', 'Dance', 'Disco',
     'Funk', 'Grunge', 'Hip-Hop', 'Jazz', 'Metal', 'New Age', 'Oldies',
     'Other', 'Pop', 'R&B', 'Rap', 'Reggae', 'Rock', 'Techno',
     'Industrial', 'Alternative', 'Ska', 'Death Metal', 'Pranks',
     'Soundtrack', 'Euro-Techno', 'Ambient', 'Trip-Hop', 'Vocal',
     'Jazz+Funk', 'Fusion', 'Trance', 'Classical', 'Instrumental',
     'Acid', 'House', 'Game', 'Sound Clip', 'Gospel', 'Noise',
     'AlternRock', 'Bass', 'Soul', 'Punk', 'Space', 'Meditative',
     'Instrumental Pop', 'Instrumental Rock', 'Ethnic', 'Gothic',
     'Darkwave', 'Techno-Industrial', 'Electronic', 'Pop-Folk',
     'Eurodance', 'Dream', 'Southern Rock', 'Comedy', 'Cult', 'Gangsta',
     'Top 40', 'Christian Rap', 'Pop/Funk', 'Jungle', 'Native American',
     'Cabaret', 'New Wave', 'Psychadelic', 'Rave', 'Showtunes',
     'Trailer', 'Lo-Fi', 'Tribal', 'Acid Punk', 'Acid Jazz', 'Polka',
     'Retro', 'Musical', 'Rock & Roll', 'Hard Rock', 'Folk',
     'Folk/Rock', 'National Folk', 'Swing', 'Fast-Fusion', 'Bebob',
     'Latin', 'Revival', 'Celtic', 'Bluegrass', 'Avantgarde',
     'Gothic Rock', 'Progressive Rock', 'Psychedelic Rock',
     'Symphonic Rock', 'Slow Rock', 'Big Band', 'Chorus',
     'Easy Listening', 'Acoustic', 'Humour', 'Speech', 'Chanson',
     'Opera', 'Chamber Music', 'Sonata', 'Symphony', 'Booty Bass',
     'Primus', 'Porn Groove', 'Satire', 'Slow Jam', 'Club', 'Tango',
     'Samba', 'Folklore', 'Ballad', 'Power Ballad', 'Rhythmic Soul',
     'Freestyle', 'Duet', 'Punk Rock', 'Drum Solo', 'A capella',
     'Euro-House', 'Dance Hall', 'Goa', 'Drum & Bass', 'Club House',
     'Hardcore', 'Terror', 'Indie', 'BritPop', 'NegerPunk',
     'Polsk Punk', 'Beat', 'Christian Gangsta', 'Heavy Metal',
     'Black Metal', 'Crossover', 'Contemporary C', 'Christian Rock',
     'Merengue', 'Salsa', 'Thrash Metal', 'Anime', 'JPop', 'SynthPop'
    );
sub parse_file
{
    my ($file, $tags) = @_;
    my ($fh, $err, $header);
    if (not (defined $file && $file ne ''))
    {
	$@ = 'No file specified';
	return -1;
    }
    if (ref $file)	# filehandle passed
    {
	$fh = $file;
    }
    else
    {
	$fh = gensym;
	if (not open $fh, "< $file\0")
	{
	    $@ = "Can't open $file: $!";
	    return -1;
	}
    }
    binmode $fh;
    # Sanity check that this looks vaguely like an MP4 file
    if ((read ($fh, $header, 8) != 8) || (lc substr ($header, 4) ne 'ftyp'))
    {
	close ($fh);
	$@ = 'Not an MPEG-4 file';
	return -1;
    }
    seek $fh, 0, 0;
    $err = parse_container($fh, 0, (stat $fh)[7], $tags);
    close ($fh);
    return $err if $err;
    # MP3::Info compatibility
    $tags->{TITLE}    = $tags->{NAM}     if defined ($tags->{NAM});
    $tags->{ARTIST}   = $tags->{ART}     if defined ($tags->{ART});
    $tags->{ALBUM}    = $tags->{ALB}     if defined ($tags->{ALB});
    $tags->{YEAR}     = $tags->{DAY}     if defined ($tags->{DAY});
    $tags->{COMMENT}  = $tags->{CMT}     if defined ($tags->{CMT});
    $tags->{GENRE}    = $tags->{GNRE}    if defined ($tags->{GNRE});
    $tags->{TRACKNUM} = $tags->{TRKN}[0] if defined ($tags->{TRKN});
    # remaining get_mp4info() stuff
    $tags->{VERSION}  = 4;
    $tags->{LAYER}    = 1                if defined ($tags->{FREQUENCY});
    $tags->{BITRATE}  = int (0.5 + $tags->{SIZE} / (($tags->{MM}*60+$tags->{SS}+$tags->{MS}/1000)*128))
	if (defined($tags->{SIZE}) && defined($tags->{MS}));	# A bit bogus
    $tags->{COPYRIGHT}= 1                if defined ($tags->{CPRT});
    $tags->{ENCRYPTED}= 0                unless defined ($tags->{ENCRYPTED});
    return 0;
}
sub parse_container
{
    my ($fh, $level, $size, $tags) = @_;
    my ($end, $err);
    $level++;
    $end = (tell $fh) + $size;
    while (tell $fh < $end)
    {
	$err = parse_atom($fh, $level, $tags);
	return $err if $err;
    }
    if (tell $fh != $end)
    {
	$@ = 'Parse error';
	return -1;
    }
    return 0;
}
sub parse_atom
{
    my ($fh, $level, $tags) = @_;
    my ($header, $size, $id, $err);
    if (read ($fh, $header, 8) != 8)
    {
	$@ = 'Premature eof';
	return -1;
    }
    ($size,$id) = unpack 'Na4', $header;
    if ($size == 1)
    {
	# extended size
	my ($hi, $lo);
	if (read ($fh, $header, 8) != 8)
	{
	    $@ = 'Premature eof';
	    return -1;
	}
	($hi,$lo) = unpack 'NN', $header;
	$size=$hi*(2**32) + $lo - 16;
    }
    else
    {
	$size -= 8;
    }
    if ($size<=0)
    {
	# Special zero-sized atom means we're done
	($size==0) && ($level==1) && return 0;
	$@ = 'Parse error';
	return -1;
    }
    $id =~ s/[^\w\-]//;
    $id = uc $id;
    printf "%s%s: %d bytes\n", ' 'x(2*$level), $id, $size if $debug;
    if (defined($data_atoms{$id}))
    {
	return parse_data ($fh, $level, $size, $id, $tags);
    }
    elsif (defined($other_atoms{$id}))
    {
	return &{$other_atoms{$id}}($fh, $level, $size, $tags);
    }
    elsif ($container_atoms{$id})
    {
	return parse_container ($fh, $level, $size, $tags);
    }
    # Unkown atom - skip past it
    seek $fh, $size, 1;
    return 0;
}
sub parse_mdat
{
    my ($fh, $level, $size, $tags) = @_;
    $tags->{SIZE} = 0 unless defined($tags->{SIZE});
    $tags->{SIZE} += $size;
    seek $fh, $size, 1;
    return 0;
}
sub parse_meta
{
    my ($fh, $level, $size, $tags) = @_;
    # META is just a container preceded by a version field
    seek $fh, 4, 1;
    return parse_container ($fh, $level, $size-4, $tags);
}
sub parse_mvhd
{
    my ($fh, $level, $size, $tags) = @_;
    my ($data, $version, $scale, $duration, $secs);
    if ($size < 32)
    {
	$@ = 'Parse error';
	return -1;
    }
    if (read ($fh, $data, $size) != $size)
    {
	$@ = 'Premature eof';
	return -1;
    }
    $version = unpack('C', $data) & 255;
    if ($version==0)
    {
	($scale,$duration) = unpack 'NN', substr ($data, 12, 8);
    }
    elsif ($version==1)
    {
	my ($hi,$lo);
	print "Long version\n" if $debug;
	($scale,$hi,$lo) = unpack 'NNN', substr ($data, 20, 12);
	$duration=$hi*(2**32) + $lo;
    }
    else
    {
	return 0;
    }
    printf "  %sDur/Scl=$duration/$scale\n", ' 'x(2*$level) if $debug;
    $secs=$duration/$scale;
    $tags->{SECS} = int (0.5+$secs);
    $tags->{MM}   = int ($secs/60);
    $tags->{SS}   = int ($secs - $tags->{MM}*60);
    $tags->{MS}   = int (0.5 + 1000*($secs - int ($secs)));
    $tags->{TIME} = sprintf "%02d:%02d",
	$tags->{MM}, $tags->{SECS} - $tags->{MM}*60;
    return 0;
}
sub parse_stsd
{
    my ($fh, $level, $size, $tags) = @_;
    my ($data, $data_format);
    if ($size < 44)
    {
	$@ = 'Parse error';
	return -1;
    }
    if (read ($fh, $data, $size) != $size)
    {
	$@ = 'Premature eof';
	return -1;
    }
    # Assumes first entry in table contains the data
    printf "  %sSample=%s\n", ' 'x(2*$level), substr ($data, 12, 4) if $debug;
    $data_format = lc substr ($data, 12, 4);
    # Is this an audio track? (Ought to look for presence of an SMHD uncle
    # atom instead to allow for other audio data formats).
    if (($data_format eq 'mp4a') ||	# AAC, aacPlus
	($data_format eq 'drms') ||	# Apple encryped AAC
	($data_format eq 'samr') ||	# Narrow-band AMR
	($data_format eq 'sawb') ||	# AMR wide-band
	($data_format eq 'sawp') ||	# AMR wide-band +
	($data_format eq 'enca'))	# Generic encryped audio
    {
#	$version = unpack "n", substr ($data, 24, 2);
#       s8.16 is inconsistent. In practice, channels always appears == 2.
#	$tags->{STEREO}  = (unpack ("n", substr ($data, 32, 2))  >  1) ? 1 : 0;
#       Old Quicktime field. No longer used.
#	$tags->{VBR}     = (unpack ("n", substr ($data, 36, 2)) == -2) ? 1 : 0;
	$tags->{FREQUENCY} = unpack ('N', substr ($data, 40, 4)) / 65536000;
	printf "  %sFreq=%s\n", ' 'x(2*$level), $tags->{FREQUENCY} if $debug;
    }
    if (($data_format eq 'drms') ||
	(substr($data_format, 0, 3) eq 'enc'))
    {
	$tags->{ENCRYPTED}=1
    }
    return 0;
}
sub parse_data
{
    my ($fh, $level, $size, $id, $tags) = @_;
    my ($data, $atom, $type);
    if (read ($fh, $data, $size) != $size)
    {
	$@ = 'Premature eof';
	return -1;
    }
    # 3GPP - different format when child of 'udta'
    if (($id eq 'TITL') ||
	($id eq 'DSCP') ||
	($id eq 'CPRT') ||
	($id eq 'PERF') ||
	($id eq 'AUTH') ||
	($id eq 'GNRE'))
    {
	my ($ver) = unpack 'N', $data;
	if ($ver == 0)
	{
	    ($size > 7) || return 0;
	    $size -= 7;
	    $type = 1;
	    $data = substr ($data, 6, $size);
	    if ($id eq 'TITL')
	    {
		return 0 if defined ($tags->{NAM});
		$id = 'NAM';
	    }
	    elsif ($id eq 'DSCP')
	    {
		return 0 if defined ($tags->{CMT});
		$id = 'CMT';
	    }
	    elsif ($id eq 'PERF')
	    {
		return 0 if defined ($tags->{ART});
		$id = 'ART';
	    }
	    elsif ($id eq 'AUTH')
	    {
		return 0 if defined ($tags->{WRT});
		$id = 'WRT';
	    }
	}
    }
    if (!defined($type))
    {
	($size > 16) || return 0;
	# Assumes first atom is the data atom we're after
	($size,$atom,$type) = unpack 'Na4N', $data;
	(lc $atom eq 'data') || return 0;
	($size > 16) || return 0;
	$size -= 16;
	$type &= 255;
	$data = substr ($data, 16, $size);
    }
    printf "  %sType=$type, Size=$size\n", ' 'x(2*$level) if $debug;
    if ($type==0)	# 16bit int data
    {
	my @ints = unpack 'n' x ($size / 2), $data;
	if ($id eq 'GNRE')
	{
	    $tags->{$id} = $mp4_genres[$ints[0]];
	}
	elsif ($size>=6)
	{
	    # Not sure what's going on here
	    $tags->{$id} = [$ints[1], $ints[2]];
	}
	else
	{
	    # Or here
	    $tags->{$id} = $ints[1];
	}
    }
    elsif ($type==1)	# Char data
    {
	# faac 1.24 and Real 10.0 encode data as unspecified 8 bit, which
	# goes against s8.28 of ISO/IEC 14496-12:2004. How tedious.
	# Assume data is utf8 if it could be utf8, otherwise assume latin1.
	if ($enable_encode) {
		my $decoder = Encode::Guess->guess ($data);
		$data = (ref ($decoder)) ?
		    $decoder->decode($data) :	# found one of utf8, utf16, latin1
		    decode($utf8 ? 'utf8' : 'latin1', $data);	# ambiguous so force
	} else {
		# taken from MP3::Info, it's non-unicode
		my $pat;
		if ($data =~ s/^\xFF\xFE//) {
			$pat = 'v';
		} elsif ($data =~ s/^\xFE\xFF//) {
			$pat = 'n';
		}
		if ($pat) {
			$data = pack 'C*', map {
				(chr =~ /[[:ascii:]]/ && chr =~ /[[:print:]]/)
					? $_
					: ord('?')
			} unpack "$pat*", $data;
		}
	}
	if ($id eq 'GEN')
	{
	    return 0 if defined ($tags->{GNRE});
	    $id='GNRE';
	}
	elsif ($id eq 'AART')
	{
	    return 0 if defined ($tags->{ART});
	    $id = 'ART';
	}
	elsif ($id eq 'DAY')
	{
	    $data = substr ($data, 0, 4);
	    # Real 10.0 supplies DAY=0 instead of deleting the atom if the
	    # year is not known. What's wrong with these people?
	    return 0 if $data==0;
	}
	$tags->{$id} = $data;
    }
    elsif ($type==21)	# Integer data
    {
	# Convert to an integer if of an appropriate size
	if ($size==1)
	{
	    $tags->{$id} = unpack 'C', $data;
	}
	elsif ($size==2)
	{
	    $tags->{$id} = unpack 'n', $data;
	}
	elsif ($size==4)
	{
	    $tags->{$id} = unpack 'N', $data;
	}
	elsif ($size==8)
	{
	    my ($hi,$lo);
	    ($hi,$lo) = unpack 'NN', $data;
	    $tags->{$id} = $hi*(2**32) + $lo;
	}
	else
	{
	    # Non-standard size - just return the raw data
	    $tags->{$id} = $data;
	}
    }
    elsif ($type==13 || $type==14)	# Raw. Appears that 13=jpeg, 14=other
    {
	$tags->{$id} = $data;
    }
    # Silently ignore other data types
    return 0;
}
1;
__END__
